#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#define NEED_CPU_REG_SHORTCUTS 1
#include "bochs.h"
#define LOG_THIS BX_CPU_THIS_PTR
#include "lockset.h"
#include "hash.h"
#include "eraser.h"
#include "globals.h"
#include "mydebug.h"


#define MAX_LOCKSETS ((unsigned int)100000)
#define HTABLE_SIZE ((unsigned int)4*MAX_LOCKSETS)   //overprovision for closed hashing

typedef unsigned int hval_t;

struct add_lock_entry {
    locksetidx_t oldindex;
    address_t newlock;
    locksetidx_t newindex;
    hash_elem h_elem;	/* Hash element */
};

struct remove_lock_entry {
    locksetidx_t oldindex;
    address_t oldlock;
    locksetidx_t newindex;
    hash_elem h_elem;	/* Hash element */
};


struct intersect_entry {
    locksetidx_t index1, index2, newindex;
    hash_elem h_elem;	/* Hash element */
};

unsigned add_lock_hash (const hash_elem *e, void *aux)
{
    struct add_lock_entry *h = hash_entry (e, struct add_lock_entry, h_elem);
    return hash_int ((h->oldindex << 16) | (h->newlock&0x0000ffff) );
}

bool
add_lock_less (const hash_elem *a_, const hash_elem *b_,
			void *aux)
{
    struct add_lock_entry *a = hash_entry (a_, struct add_lock_entry, h_elem);
    struct add_lock_entry *b = hash_entry (b_, struct add_lock_entry, h_elem);
    return (a->oldindex < b->oldindex || (a->oldindex==b->oldindex && a->newlock < b->newlock));
}

unsigned remove_lock_hash (const hash_elem *e, void *aux)
{
    struct remove_lock_entry *h = hash_entry (e, struct remove_lock_entry, h_elem);
    return hash_int ((h->oldindex << 16) | (h->oldlock&0x0000ffff));
}

bool
remove_lock_less (const hash_elem *a_, const hash_elem *b_,
			void *aux)
{
    struct remove_lock_entry *a = hash_entry (a_, struct remove_lock_entry, h_elem);
    struct remove_lock_entry *b = hash_entry (b_, struct remove_lock_entry, h_elem);
    return (a->oldindex < b->oldindex || (a->oldindex==b->oldindex && a->oldlock < b->oldlock));
}

unsigned intersect_hash (const hash_elem *e, void *aux)
{
    struct intersect_entry *h = hash_entry (e, struct intersect_entry, h_elem);
    return hash_int ((h->index1 << 16) | (h->index2&0x0000ffff));
}

bool
intersect_less (const hash_elem *a_, const hash_elem *b_,
			void *aux)
{
    struct intersect_entry *a = hash_entry (a_, struct intersect_entry, h_elem);
    struct intersect_entry *b = hash_entry (b_, struct intersect_entry, h_elem);
    return (a->index1 < b->index1 || (a->index1==b->index1 && a->index2 < b->index2));
}




static int add_lock_cache_initialized = 0;
static struct hash add_lock_cache;

static int remove_lock_cache_initialized = 0;
static struct hash remove_lock_cache;

static int intersect_cache_initialized = 0;
static struct hash intersect_cache;

typedef struct lockvector {
  address_t lockaddress;
  struct lockvector *next;
} lockvector_t;


static locksetidx_t lockset_index = 1;
static lockvector_t *index_table[MAX_LOCKSETS];
static locksetidx_t hash_table[HTABLE_SIZE];

static unsigned int cum_hash(unsigned int a, address_t b) {
    return (3*a+b);
}

static unsigned int hashfn(lockvector_t *vec) {
    lockvector_t *cur;
    unsigned int ret=0;
    assert(vec);
    cur = vec;
    while (cur) {
        ret = cum_hash(ret,cur->lockaddress);
	cur = cur->next;
    }
    return ret;
}

static void
freevector(lockvector_t *lockvec) {
    static lockvector_t *cur, *prev;
    cur = lockvec;
    while (cur) {
        prev = cur;
	cur = cur->next;
	free(prev);
    }
}

static lockvector_t *
clonevector(lockvector_t *lockvec) {
    static lockvector_t *cur, *newlockvec,*newprev;
    if (!lockvec) return NULL;
    assert(lockvec);
    newlockvec = (lockvector_t*)malloc(sizeof(lockvector_t));
    newlockvec->lockaddress = lockvec->lockaddress;
    newprev = newlockvec;
    cur = lockvec->next;
    while (cur) {
        newprev->next = (lockvector_t*)malloc(sizeof(lockvector_t));
	newprev->next->lockaddress = cur->lockaddress;
	cur = cur->next;
	newprev = newprev->next;
    }
    newprev->next = NULL;
    return newlockvec;
}

static lockvector_t *
remove_from_lockvector(lockvector_t *lockvec, address_t oldlock) {
    lockvector_t *ret, *cur, *newlockvec, *tmp;
    assert(lockvec);

    if (oldlock<lockvec->lockaddress) assert(0);
    if (oldlock==lockvec->lockaddress) {
	ret = clonevector(lockvec->next);
	return ret;
    }

    cur = newlockvec = clonevector(lockvec);
    while (cur->next && oldlock>cur->next->lockaddress) {
        cur = cur->next;
    }
    assert(oldlock==cur->next->lockaddress);

    tmp = cur->next;
    cur->next = cur->next->next;
    tmp->lockaddress = 0;
    tmp->next = NULL;
    free(tmp);
    return newlockvec;
}

static lockvector_t *
intersect_lockvectors(lockvector_t *lockvec1, lockvector_t *lockvec2) {
    lockvector_t *cur1, *cur2, *out, *ret;
    cur1 = lockvec1;
    cur2 = lockvec2;
    ret = out = NULL;
    while (cur1 && cur2) {
	if (cur1->lockaddress < cur2->lockaddress) {
	    cur1 = cur1->next;
	} else if (cur1->lockaddress > cur2->lockaddress) {
	    cur2 = cur2->next;
	} else {
	    if (out) {
	        out->next = (lockvector_t *)malloc(sizeof(lockvector_t));
		out = out->next;
	    } else {
	        ret = out = (lockvector_t *)malloc(sizeof(lockvector_t));
	    }
	    out->lockaddress = cur1->lockaddress;
	    out->next = NULL;
	    cur1 = cur1->next;
	    cur2 = cur2->next;
	}
    }
    return ret;
}




static lockvector_t *
add_to_lockvector(lockvector_t *lockvec, address_t newlock) {
    lockvector_t *ret, *cur, *newlockvec;
    if (lockvec==NULL) {
	ret = (lockvector_t *)malloc(sizeof(lockvector_t));
	ret->lockaddress = newlock;
	ret->next = NULL;
	return ret;
    }
    if (newlock<lockvec->lockaddress) {
	ret = (lockvector_t *)malloc(sizeof(lockvector_t));
	ret->lockaddress = newlock;
	ret->next = clonevector(lockvec);
	return ret;
    }
    if (newlock==lockvec->lockaddress) {
	if (BX_CPU(0)) BX_CPU(0)->backtrace(btstr);
        assert(0);
	return clonevector(lockvec);
    }
    assert(newlock>lockvec->lockaddress);

    cur = newlockvec = clonevector(lockvec);
    while (cur->next && newlock>cur->next->lockaddress) {
        cur = cur->next;
    }

    if (!cur->next) {
	ret = (lockvector_t *)malloc(sizeof(lockvector_t));
	ret->lockaddress = newlock;
	cur->next = ret;
	ret->next = NULL;
	return newlockvec;
    }

    if (newlock==cur->next->lockaddress) {
	BX_CPU(0)->backtrace(btstr);
	assert(0); //acquiring the same lock twice
        return newlockvec;
    }

    ret = (lockvector_t *)malloc(sizeof(lockvector_t));
    ret->lockaddress = newlock;
    ret->next = cur->next;
    cur->next = ret;
    return newlockvec;
}

void add_to_hashtable(lockvector_t *newlockset, locksetidx_t index) {
    hval_t hval = (hashfn(newlockset))%HTABLE_SIZE;
    int numtries = 0;
    while (hash_table[hval]) {
        hval = (hval+1)%HTABLE_SIZE; //closed hashing
	//printf("hval=%d.\n",hval);
	numtries++;
    }
    if (numtries>20) printf("%s: Warning: numtries too large for closed hashing (=%d).\n",__func__,numtries);
    hash_table[hval] = index;
}

//returns true if they are equal
static int
compare_locksets(lockvector_t *vec1, lockvector_t *vec2) {
    lockvector_t *cur1, *cur2;

    assert(vec1);
    assert(vec2);
    
    cur1 = vec1;
    cur2 = vec2;
    while (cur1 && cur2 && cur1->lockaddress==cur2->lockaddress) {
	cur1 = cur1->next;
	cur2 = cur2->next;
    }

    if (cur1) return 0;
    if (cur2) return 0;
    return 1;
}


static locksetidx_t
find_in_add_cache(locksetidx_t index, address_t newlock) {
    hval_t hval;
    locksetidx_t new_index;
    hash_elem *h_element = NULL;
    struct add_lock_entry tmp;
    tmp.oldindex = index; tmp.newlock = newlock;
    if (add_lock_cache_initialized) h_element = hash_find(&add_lock_cache, &tmp.h_elem);
    if (h_element) {
        struct add_lock_entry *answer = hash_entry(h_element, struct add_lock_entry, h_elem);
	//printf("%s: returning (%d, %x)->%d from cache.\n",__func__,index,newlock,answer->newindex);
	return answer->newindex;
    }
    return 0;
}


static locksetidx_t
find_in_remove_cache(locksetidx_t index, address_t oldlock) {
    hval_t hval;
    locksetidx_t new_index;
    hash_elem *h_element = NULL;
    struct remove_lock_entry tmp;
    tmp.oldindex = index; tmp.oldlock = oldlock;
    if (remove_lock_cache_initialized) h_element = hash_find(&remove_lock_cache, &tmp.h_elem);
    if (h_element) {
        struct remove_lock_entry *answer = hash_entry(h_element, struct remove_lock_entry, h_elem);
	//printf("%s: returning (%d, %d)->%d from cache.\n",__func__,index,oldlock,answer->newindex);
	return answer->newindex;
    }

    return 0;
}

static locksetidx_t
find_in_intersect_cache(locksetidx_t index1, locksetidx_t index2) {
    hval_t hval;
    locksetidx_t new_index;
    hash_elem *h_element = NULL;
    struct intersect_entry tmp;
    tmp.index1 = index1; tmp.index2 = index2;
    if (intersect_cache_initialized) h_element = hash_find(&intersect_cache, &tmp.h_elem);
    if (h_element) {
        struct intersect_entry *answer = hash_entry(h_element, struct intersect_entry, h_elem);
	//printf("%s: returning (%d, %x)->%d from cache.\n",__func__,index1,index2,answer->newindex);
	return answer->newindex;
    }
    return 0;
}


void insert_in_remove_cache(locksetidx_t index, address_t lock, locksetidx_t new_index)
{
    struct remove_lock_entry *e;
    if (!remove_lock_cache_initialized) {
	hash_init(&remove_lock_cache, remove_lock_hash, remove_lock_less, NULL);
	remove_lock_cache_initialized = 1;
    }
    e = (struct remove_lock_entry *)malloc(sizeof(struct remove_lock_entry));
    e->oldindex = index; e->oldlock = lock; e->newindex = new_index;
    hash_insert(&remove_lock_cache, &e->h_elem);
}

void
insert_in_add_cache(locksetidx_t index, address_t lock, locksetidx_t new_index)
{
    struct add_lock_entry *e;
    if (!add_lock_cache_initialized) {
	hash_init(&add_lock_cache, add_lock_hash, add_lock_less, NULL);
	add_lock_cache_initialized = 1;
    }
    e = (struct add_lock_entry *)malloc(sizeof(struct add_lock_entry));
    e->oldindex = index; e->newlock = lock; e->newindex = new_index;
    hash_insert(&add_lock_cache, &e->h_elem);
}

void
insert_in_intersect_cache(locksetidx_t index1, locksetidx_t index2, locksetidx_t new_index)
{
    struct intersect_entry *e;
    if (!intersect_cache_initialized) {
	hash_init(&intersect_cache, intersect_hash, intersect_less, NULL);
	intersect_cache_initialized = 1;
    }
    e = (struct intersect_entry *)malloc(sizeof(struct intersect_entry));
    e->index1 = index1; e->index2 = index2; e->newindex = new_index;
    hash_insert(&intersect_cache, &e->h_elem);
}

static locksetidx_t
find_duplicate(lockvector_t *vec) {
    hval_t hval;
    locksetidx_t new_index;
    hash_elem *h_element = NULL;

    hval = (hashfn(vec))%HTABLE_SIZE;
    new_index = hash_table[hval];

    while (new_index) {
        if (compare_locksets(vec,index_table[new_index])) {
	  return new_index;
	}
	hval = (hval+1)%HTABLE_SIZE;
	new_index = hash_table[hval];
    }
    return 0;
}

locksetidx_t add_lock(locksetidx_t index, address_t newlock) {
    lockvector_t *newlockset;
    locksetidx_t ret_index;
    ret_index = find_in_add_cache(index,newlock);
    if (ret_index) return ret_index;

    if (lockset_index >= MAX_LOCKSETS) {
	printf("%s: Number of locksets exceed %d. returning 0\n",__func__,MAX_LOCKSETS);
        return 0;
    }

    if (!index) {	//adding a lock to an empty lockset
        newlockset = (lockvector_t *)malloc(sizeof(lockvector_t));
	newlockset->lockaddress = newlock;
	newlockset->next = NULL;
    } else {
	newlockset = add_to_lockvector(index_table[index],newlock);
    }
    assert(newlockset);
    ret_index = find_duplicate(newlockset);
    if (ret_index) {
        insert_in_add_cache(index,newlock,ret_index);
        freevector(newlockset);
	return ret_index;
    }

    index_table[lockset_index] = newlockset;
    add_to_hashtable(newlockset,lockset_index);
    ret_index = lockset_index;
    lockset_index++;
    return ret_index;
}

static int
singleton_lockset(lockvector_t *vec) {
    assert(vec);
    if (vec->next==NULL) return 1;
    return 0;
}

locksetidx_t remove_lock(locksetidx_t index, address_t oldlock) {
    lockvector_t *newlockset;
    locksetidx_t ret_index;
    if (!index_table[index]) {
	assert(oldlock==INTERRUPT_LOCK);
	assert(index==0);
	return 0;
    }
    if (singleton_lockset(index_table[index])) {
        if (index_table[index]->lockaddress!=oldlock) {
	    BX_CPU(0)->backtrace(btstr);
	    printf("ERROR: index=%d, index_table[index]->lockaddress=%x, oldlock=%x.\n",index,index_table[index]->lockaddress,oldlock);
	}
        assert(index_table[index]->lockaddress==oldlock);
	return 0;
    }
    ret_index = find_in_remove_cache(index,oldlock);
    if (ret_index) return ret_index;

    if (lockset_index >= MAX_LOCKSETS) {
	printf("%s: Number of locksets exceed %d. returning 0\n",__func__,MAX_LOCKSETS);
        return 0;
    }

    assert(index);	//you cannot remove from an empty lockset
    newlockset = remove_from_lockvector(index_table[index],oldlock);
    assert(newlockset);

    ret_index = find_duplicate(newlockset);
    if (ret_index) {
        insert_in_remove_cache(index,oldlock,ret_index);
        freevector(newlockset);
	return ret_index;
    }

    index_table[lockset_index] = newlockset;
    add_to_hashtable(newlockset,lockset_index);
    ret_index = lockset_index;
    lockset_index++;
    return ret_index;
}

locksetidx_t intersect_locksets(locksetidx_t index1, locksetidx_t index2) {
    lockvector_t *newlockset;
    locksetidx_t ret_index;

    if (index1==0) return 0;
    if (index2==0) return 0;
    
    ret_index = find_in_intersect_cache(index1,index2);
    if (ret_index) return ret_index;

    if (lockset_index >= MAX_LOCKSETS) {
	printf("%s: Number of locksets exceed %d. returning 0\n",__func__,MAX_LOCKSETS);
        return 0;
    }
    assert(index1);
    assert(index2);

    newlockset = intersect_lockvectors(index_table[index1],index_table[index2]);
    if (!newlockset) return 0;		//empty lockset
    ret_index = find_duplicate(newlockset);
    if (ret_index) {
        insert_in_intersect_cache(index1,index2,ret_index);
        freevector(newlockset);
	return ret_index;
    }

    index_table[lockset_index] = newlockset;
    add_to_hashtable(newlockset,lockset_index);
    ret_index = lockset_index;
    lockset_index++;
    return ret_index;
}

void lockset_dump(locksetidx_t idx, Bit32u *locks, int cnt) 
{
    lockvector_t *vec = index_table[idx];
    while (vec != NULL && cnt-- > 0) {
        *locks++ = vec->lockaddress;
        vec = vec->next;
    }
    while (cnt-- > 0) {
        *locks++ = 0;
    }
}

typedef struct locks_held {
    unsigned threadId;
    locksetidx_t locks_held;
    int ignore;	//whether this thread should be ignored. =0 if not. otherwise stores ignore depth
    hash_elem h_elem;	/* Hash element */
} locks_held_t ;

static int locks_held_table_initialized = 0;
static struct hash locks_held_table ;

unsigned locks_held_hash (const hash_elem *e, void *aux)
{
    locks_held_t *h = hash_entry (e, struct locks_held, h_elem);
    return hash_int (h->threadId);
}

bool
locks_held_less (const hash_elem *a_, const hash_elem *b_,
			void *aux)
{
    locks_held_t *a = hash_entry (a_, struct locks_held, h_elem);
    locks_held_t *b = hash_entry (b_, struct locks_held, h_elem);
    return (a->threadId < b->threadId);
}


locksetidx_t
cur_held(unsigned threadId) {
    hash_elem *h_element;
    locks_held_t tmp;
    if (!locks_held_table_initialized) return 0;
    tmp.threadId = threadId;
    h_element = hash_find(&locks_held_table, &tmp.h_elem);
    if (h_element) {
        locks_held_t *answer = hash_entry(h_element, struct locks_held, h_elem);
	assert(answer->locks_held<lockset_index);
	return answer->locks_held;
    }
    return 0;
}

void update_lockset(unsigned threadId, locksetidx_t lset) {
    locks_held_t *e;
    hash_elem *h_element;
    locks_held_t tmp;
    assert(lset<lockset_index);
    if (!locks_held_table_initialized) {
	hash_init(&locks_held_table, locks_held_hash, locks_held_less, NULL);
	locks_held_table_initialized = 1;
    }
    tmp.threadId = threadId;
    h_element = hash_find(&locks_held_table, &tmp.h_elem);
    if (!h_element) {
	e = (struct locks_held *)malloc(sizeof(struct locks_held));
	e->threadId = threadId; e->locks_held = lset; e->ignore = 0;
	hash_insert(&locks_held_table, &e->h_elem);
    } else {
	e = hash_entry(h_element, struct locks_held, h_elem);
	assert(e->threadId==threadId);
	e->locks_held = lset;
    }
}


static bool
find_in_lockvector(lockvector_t *lockvec, address_t lock) {
    lockvector_t *cur;
    assert(lockvec!=NULL);
    if (lock<lockvec->lockaddress) return false;
    if (lock==lockvec->lockaddress) return true;
    cur = lockvec;
    while (cur->next && lock>cur->next->lockaddress) {
        cur = cur->next;
    }
    if (!cur->next) return false;
    if (cur->next->lockaddress==lock) return true;
    return false;
}

bool belongs(locksetidx_t index, address_t lock) {
    if (!index) return false;
    return (find_in_lockvector(index_table[index],lock));
}

void set_ignore(unsigned threadId, bool val) {
    locks_held_t *e;
    hash_elem *h_element;
    locks_held_t tmp;
    if (!locks_held_table_initialized) {
	hash_init(&locks_held_table, locks_held_hash, locks_held_less, NULL);
	locks_held_table_initialized = 1;
    }
    tmp.threadId = threadId;
    h_element = hash_find(&locks_held_table, &tmp.h_elem);
    if (!h_element) {
	e = (struct locks_held *)malloc(sizeof(struct locks_held));
	e->threadId = threadId; e->locks_held = LOCKSET_EMPTY; e->ignore=0;
	if (val) e->ignore++; else if (e->ignore) e->ignore--;
	assert(e->ignore>=0);
	hash_insert(&locks_held_table, &e->h_elem);
    } else {
	e = hash_entry(h_element, struct locks_held, h_elem);
	assert(e->threadId==threadId);
	if (val) e->ignore++; else if (e->ignore) e->ignore--;
	assert(e->ignore>=0);
    }
    DBG(L1,("ignore value for thread %x = %d.\n",threadId,e->ignore));
}

bool
ignore_on(unsigned threadId) {
    hash_elem *h_element;
    locks_held_t tmp;
    if (global_startup_ignore) return true;
    if (!locks_held_table_initialized) return 0;
    tmp.threadId = threadId;
    h_element = hash_find(&locks_held_table, &tmp.h_elem);
    if (h_element) {
        locks_held_t *answer = hash_entry(h_element, struct locks_held, h_elem);
	assert(answer->ignore>=0);
	return answer->ignore > 0;
    }
    return false;
}
